!> module for the operation with FE solution 
module o_sol
  use o_mesh
  use o_matrix
  
  implicit none

contains

  ! compute the error in the L^2-norm of the P1 FE solution 
  subroutine ComputeError(grid,  u, err )
    type(mesh), target, intent(inout) :: grid
    real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    real, intent(inout) :: err
    real, dimension(:,:), pointer :: xp  ! nodes coordinates
    integer, dimension(:,:), pointer :: nodes  ! index vertex
    integer :: i,j,k, l, npoin, nelem, Qdof
    real :: err_loc, x, y
    real, dimension(:), allocatable :: weights, uh, uE
    real, dimension(:,:), allocatable :: lambda
    
    
    !  quadrature rule
    Qdof = 6
    allocate(weights(1:Qdof), lambda(1:Qdof, 3) )
    weights(  1) =  2.2338158967801100E-01
    weights(  2) =  2.2338158967801100E-01
    weights(  3) =  2.2338158967801100E-01
    weights(  4) =  1.0995174365532200E-01
    weights(  5) =  1.0995174365532200E-01
    weights(  6) =  1.0995174365532200E-01
    
    lambda(  1,1:3) = (/ 1.0810301816807000E-01,  4.4594849091596500E-01,  4.4594849091596500E-01/)
    lambda(  2,1:3) = (/ 4.4594849091596500E-01,  4.4594849091596500E-01,  1.0810301816807000E-01/)
    lambda(  3,1:3) = (/ 4.4594849091596500E-01,  1.0810301816807000E-01,  4.4594849091596500E-01/)
    lambda(  4,1:3) = (/ 8.1684757298045896E-01,  9.1576213509771007E-02,  9.1576213509770035E-02/)
    lambda(  5,1:3) = (/ 9.1576213509771007E-02,  9.1576213509771007E-02,  8.1684757298045796E-01/)
    lambda(  6,1:3) = (/ 9.1576213509771007E-02,  8.1684757298045896E-01,  9.1576213509770035E-02/)
    !  quadrature rule

    allocate(uh(1:Qdof), uE(1:Qdof))
    
    npoin = grid%npoin
    nelem = grid%nelem

    nodes => grid%nodes(1:nelem,1:3)

    xp => grid%xp(1:npoin,1:2)

    err = 0

    do i=1,grid%nelem
       err_loc = 0.
       
       do j=1,Qdof
          ! quadrature node
          x  =  dot_product( xp(nodes(i, 1:3), 1) , lambda(j, 1:3))
          y  =  dot_product( xp(nodes(i, 1:3), 2) , lambda(j, 1:3))
          uE(j) =  set_DBC(x, y)
          uh(j) =   dot_product( u(nodes(i,1:3)), lambda(j, 1:3) )

          !write(111,*) x,y,uh(j), uE(j), uh(j) - uE(j)

          err_loc = err_loc + grid%area(i) * weights(j) * (uh(j) - uE(j))**2
       enddo
       err = err + err_loc
    enddo

    err = sqrt(err)
  end subroutine ComputeError

    ! compute the error in the H^1-norm of the P1 FE solution 
  subroutine ComputeErrorH1(grid,  u, err )
    type(mesh), target, intent(inout) :: grid
    real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    real, intent(inout) :: err
    real, dimension(:,:), allocatable :: Dphi, xi
    real, dimension(:,:), pointer :: xp  ! nodes coordinates
    integer, dimension(:,:), pointer :: nodes  ! index vertex
    integer :: i,j, npoin, nelem, Qdof
    real :: err_loc,  D
    real, dimension(:), allocatable :: uh, uE, nabla_E
    
    !  quadrature rule woth vertices
    Qdof = 3 ! 
    allocate(uh(1:Qdof), uE(1:Qdof))
    allocate(nabla_E(1:2) )
    allocate(Dphi(1:3, 1:2), xi(1:3, 1:2) )

    npoin = grid%npoin
    nelem = grid%nelem

    nodes => grid%nodes(1:nelem,1:3)

    xp => grid%xp(1:npoin,1:2)

    err = 0

    do i=1,grid%nelem
       ! coordinates of the element
       xi(1:3, 1:2) =  xp(nodes(i, 1:3), 1:2 )

       ! double area of element
       D =  2 * grid%area(i) 

       call GRADIENTS( xi(1:3,1), xi(1:3, 2), Dphi(1:3, 1:2) )
       Dphi(1:3, 1:2) = Dphi(1:3, 1:2) / D
       
       nabla_E = 0.
       do j=1,Qdof
          ! quadrature node
          uE(j) =  set_DBC(xi(j,1), xi(j,2) )
          uh(j) =  u(nodes(i,j))
          !write(111,*) x,y,uh(j), uE(j), uh(j) - uE(j)

          nabla_E(1) = nabla_E(1)  + (uh(j) - uE(j)) * Dphi(j,1)
          nabla_E(2) = nabla_E(2)  + (uh(j) - uE(j)) * Dphi(j,2)
       enddo


       err_loc =  grid%area(i) * dot_product(nabla_E(1:2), nabla_E(1:2) ) / 3
       err = err + err_loc 
       !write(*,'(i5,30es12.4)') i, xi(1,1:2),nabla_E, err_loc, err, dot_product(nabla_E(1:2), nabla_E(1:2) )

    enddo

    err = sqrt(err)
    deallocate(nabla_E, uE, uh)
    
  end subroutine ComputeErrorH1

  ! plot of the P1 FE solution to a file suitbale for gnuplot
  subroutine Plot_solution(grid,  solfile, u )
    type(mesh), target, intent(inout) :: grid
   character(len=*), intent( in) :: solfile
   real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    real, dimension(:,:), pointer :: xp  ! nodes coordinates
    integer, dimension(:,:), pointer :: nodes  ! index vertex
    integer :: i,j,k, l, npoin, nelem

    npoin = grid%npoin
    nelem = grid%nelem

    nodes => grid%nodes(1:nelem,1:3)

    xp => grid%xp(1:npoin,1:2)

    open(90, file=solfile, status='UNKNOWN')

  ! drawing of matrix structure
    do i=1,grid%nelem

       write(90, *) xp(nodes(i, 1), 1:2), u(nodes(i,1))
       write(90, *) xp(nodes(i, 2), 1:2), u(nodes(i,2))
       write(90, *) xp(nodes(i, 3), 1:2), u(nodes(i,3))
       write(90, *) xp(nodes(i, 1), 1:2), u(nodes(i,1))
       write(90, '(x)' )
       write(90, '(x)' )
       write(90, '(x)' )
    enddo

    close(90)
    
  end subroutine Plot_solution

  ! write of the P1 FE solution to a file for ANGENER
  subroutine Write_solution(grid,  solfile, u )
    type(mesh), target, intent(inout) :: grid
    character(len=*), intent( in) :: solfile
    real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    !real, dimension(:,:), pointer :: xp  ! nodes coordinates
    !integer, dimension(:,:), pointer :: nodes  ! index vertex
    integer :: i,j,k, l, npoin, nelem

    npoin = grid%npoin
    nelem = grid%nelem

    !nodes => grid%nodes(1:nelem,1:3)

    !xp => grid%xp(1:npoin,1:2)

    open(90, file=solfile, status='UNKNOWN')

  ! drawing of matrix structure
    do i=1,grid%npoin
       write(90, *) u(i)
    enddo

    close(90)
    
  end subroutine Write_solution

  ! setting of the RHS for the linear algebraci problem
  ! source term + terms from non-homogeous Dirichlet BC
  subroutine InitSol_RHS(grid, mS, u,  v,f)
    type(mesh), target, intent(inout) :: grid
    real, dimension(1:grid%npoin, 1:grid%npoin), intent(inout) :: mS  ! stiff matrix
    real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    real, dimension(1:grid%ipoin), intent(inout) :: v  ! init solution in nodes
    real, dimension(1:grid%ipoin), intent(inout) :: f  ! RHS in interior nodes
    real, dimension(:,:), pointer :: xp  ! nodes coordinates
    real, dimension(:), pointer :: area  ! area of elements
    integer, dimension(:,:), pointer :: nodes  ! vertex indexes
    integer, dimension(:,:), pointer :: ip  ! interior / boundary
    real, dimension(:), allocatable :: xi  ! coordinates in middle of edges
    integer :: i,j, j1, k, l, npoin, ipoin, nelem, k1, k2
    real :: val

    allocate(xi(1:2) )
    
    npoin = grid%npoin
    ipoin = grid%ipoin
    nelem = grid%nelem

    ip => grid%ip(1:npoin,1:2)

    xp => grid%xp(1:npoin,1:2)

    nodes => grid%nodes(1:nelem,1:3)

    area => grid%area(1:nelem)

    
    u = 0.

    ! setting of the Dicihlet BC
    do i=1, npoin
       if(ip(i, 1) < 0) then
          u(i) = set_DBC(xp(i, 1), xp(i,2) )
       else
          ! initial setting of the solution in inner and Neumann nodes
          v(ip(i,2) ) = u(i) ! usually =0
       endif
    enddo

    ! setting of source term
    f(:) = 0
    
    do i=1, nelem
       do j=1, 3
          j1 = mod(j, 3) + 1

          k1 = nodes(i, j)
          k2 = nodes(i, j1)
          xi(1:2) = ( xp( k1, 1:2) + xp(k2, 1:2) ) /2

          val = set_RHS( xi(1), xi(2) )

          ! Neumann or interior nodes
          if(ip(k1, 1) >= 0) f(ip(k1,2)) = f(ip(k1,2)) + val * area(i) / 6
          if(ip(k2, 1) >= 0) f(ip(k2,2)) = f(ip(k2,2)) + val * area(i) / 6
       enddo
       
    enddo


    ! modification  of the right-hand side by the Dirichlet BC
    do i=1, npoin
       if(ip(i, 1) >= 0) then ! interior node, the equation
          k = ip(i,2)

          do j=1, npoin
             if(ip(j, 1) < 0) then  ! Dirichlet  node (the data)
                
                f(k) = f(k) - mS(i, j) * u(j )   ! u already contains Dirichlet data
             endif
          enddo
       endif
    end do


    ! modification  of the right-hand side by the Neumann BC
    do i=1, npoin
       if(ip(i, 1) == 0) then ! Neumann  node, the equation
          k = ip(i,2)

          val = set_NBC(xp(k, 1), xp(k,2) )
          
          f(k) = f(k) + val * grid%b_edge(i) 
       endif
    end do

    
    deallocate(xi )
    
  end subroutine InitSol_RHS



  
  ! setting of the RHS for the linear algebraci problem
  ! source term + terms from non-homogeous Dirichlet BC
  subroutine InitSol_RHS_Sparse(grid, u,  v,f)
    type(mesh), target, intent(inout) :: grid
    !!real, dimension(1:grid%npoin, 1:grid%npoin), intent(inout) :: mS  ! stiff matrix
    real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    real, dimension(1:grid%ipoin), intent(inout) :: v  ! init solution in nodes
    real, dimension(1:grid%ipoin), intent(inout) :: f  ! RHS in interior nodes
    real, dimension(:,:), pointer :: xp  ! nodes coordinates
    real, dimension(:), pointer :: area  ! area of elements
    integer, dimension(:,:), pointer :: nodes  ! vertex indexes
    integer, dimension(:,:), pointer :: ip  ! interior / boundary
    real, dimension(:), allocatable :: xi  ! coordinates in middle of edges
    integer :: i,j, j1, k, l, js, npoin, ipoin, nelem, k1, k2, kk, nzero
    real :: val

    allocate(xi(1:2) )
    
    npoin = grid%npoin
    ipoin = grid%ipoin
    nelem = grid%nelem

    nzero = grid%nzero

    ip => grid%ip(1:npoin,1:2)

    xp => grid%xp(1:npoin,1:2)

    nodes => grid%nodes(1:nelem,1:3)

    area => grid%area(1:nelem)

    
    u = 0.

    ! setting of the Dicihlet BC
    do i=1, npoin
       if(ip(i, 1) < 0) then
          u(i) = set_DBC(xp(i, 1), xp(i,2) )
       else
          ! initial setting of the solution in inner and Neumann nodes
          v(ip(i,2) ) = u(i) ! usually =0
       endif
    enddo

    ! setting of source term
    f(:) = 0
    
    do i=1, nelem
       do j=1, 3
          j1 = mod(j, 3) + 1

          k1 = nodes(i, j)
          k2 = nodes(i, j1)
          xi(1:2) = ( xp( k1, 1:2) + xp(k2, 1:2) ) /2

          val = set_RHS( xi(1), xi(2) )

          ! Neumann or interior nodes
          if(ip(k1, 1) >= 0) f(ip(k1,2)) = f(ip(k1,2)) + val * area(i) / 6
          if(ip(k2, 1) >= 0) f(ip(k2,2)) = f(ip(k2,2)) + val * area(i) / 6
       enddo
       
    enddo


    ! modification  of the right-hand side by the Dirichlet BC
    do i=1, npoin
       if(ip(i, 1) >= 0) then ! interior node, the equation
          k = ip(i,2)

          do js = grid%irow(i), grid%irow(i+1) - 1
             j = grid%icol(js)
             
             if(ip(j, 1) < 0) then  ! Dirichlet  node (the data)
                
                kk = Find_Sparce_index(npoin, nzero, grid%irow, grid%icol, i, j)
                
                !f(k) = f(k) - mS(i, j) * u(j )   ! u already contains Dirichlet data
                f(k) = f(k) - grid%sparse(kk) * u(j )   ! u already contains Dirichlet data
             endif
          enddo
       endif
    end do


    ! modification  of the right-hand side by the Neumann BC
    do i=1, npoin
       if(ip(i, 1) == 0) then ! Neumann  node, the equation
          k = ip(i,2)

          val = set_NBC(xp(k, 1), xp(k,2) )
          
          f(k) = f(k) + val * grid%b_edge(i) 
       endif
    end do

    
    deallocate(xi )
    
  end subroutine InitSol_RHS_Sparse


  ! setting of the Dirichlet boundary condition at x,y
  function set_DBC(x, y) result( fun )
    real, intent(in) :: x, y  ! point coordinates
    real :: fun

    !fun = 0.
    fun = x**2 + y**2
    !fun = (1 - x**10) * (1-y**20) 

  end function set_DBC


  ! setting of the Neumann boundary condition at x,y
  function set_NBC(x, y) result( fun )
    real, intent(in) :: x, y  ! point coordinates
    real :: fun

    fun = 0.
    fun = 1.

  end function set_NBC


  ! evaluation of the RHS of the equation at x,y
  function set_RHS(x, y) result( fun )
    real, intent(in) :: x, y  ! point coordinates
    real :: fun

    !fun = 2.
    !fun = 2*y*(1-y) + 2*x*(1-x)
    fun = -4

    !fun = 90* x**8 *(1 - y**20)  + 380 * y**18 *(1-x**10) 
    
  end function set_RHS

end module o_sol


